import math
import os
from dataclasses import asdict, dataclass
from typing import Dict, List, Literal, Optional, Tuple

import hyperstate
import numpy as np
import numpy.typing as npt
import torch
import torch.nn as nn
import torch.nn.functional as F
import wandb
from entity_gym.env.vec_env import VecActionMask
from entity_gym.ragged_dict import RaggedActionDict, RaggedBatchDict
from entity_gym.serialization import Trace
from entity_gym.serialization.sample_loader import Episode, MergedSamples
from entity_gym.simple_trace import Tracer
from ragged_buffer import RaggedBufferBool, RaggedBufferF32, RaggedBufferI64
from rogue_net.rogue_net import RogueNet, RogueNetConfig
from torch.optim import AdamW


@dataclass
class OptimizerConfig:
    """Optimizer hyperparameters

    Attributes:
        lr: learning rate
        anneal_lr: anneal learning rate
        max_grad_norm: max gradient norm
        batch_size: batch size
    """

    lr: float = 1e-4
    anneal_lr: bool = True
    max_grad_norm: float = 100.0
    batch_size: int = 512


@dataclass
class WandbConfig:
    """W&B tracking settings.

    Args:
        track: whether to track metrics to W&B
        project_name: the wandb's project name
        entity: the entity (team) of wandb's project
    """

    track: bool = False
    project_name: str = "enn-bc"
    entity: str = "entity-neural-network"


@dataclass
class Config:
    """Supervised training configuration.

    Args:
        optim: optimizer configuration
        wandb: wandb tracking settings
        model: transformer network hyperparameters
        dataset_path: file to load training/test dataset from
        epochs: number of epochs to train for
        loss_fn: loss function ("kl" or "mse")
        log_interval: print out loss every log_interval steps
        fast_eval_interval: interval at which to evaluate with subset of test data
        fast_eval_samples: number of samples to use in fast evaluation
    """

    optim: OptimizerConfig
    wandb: WandbConfig
    model: RogueNetConfig
    dataset_path: str
    epochs: int = 10
    loss_fn: Literal["kl", "mse"] = "mse"
    log_interval: int = 10
    fast_eval_interval: int = 32768
    fast_eval_samples: int = 8192


@dataclass
class DataSet:
    entities: RaggedBatchDict[np.float32]
    visible: RaggedBatchDict[np.bool_]
    actions: RaggedBatchDict[np.int64]
    logprobs: RaggedBatchDict[np.float32]
    masks: RaggedActionDict
    logits: Optional[RaggedBatchDict[np.float32]]
    batch_size: int
    frames: int

    permutation: Optional[npt.NDArray[np.int64]] = None

    @classmethod
    def from_merged_samples(
        cls, merged_samples: MergedSamples, batch_size: int
    ) -> "DataSet":
        frames = (merged_samples.frames // batch_size) * batch_size
        if frames == 0:
            frames = merged_samples.frames
            batch_size = frames
        return cls(
            entities=merged_samples.entities,
            visible=merged_samples.visible,
            actions=merged_samples.actions,
            logprobs=merged_samples.logprobs,
            masks=merged_samples.masks,
            logits=merged_samples.logits,
            batch_size=batch_size,
            frames=frames,
        )

    @classmethod
    def from_episodes(cls, episodes: List[Episode], batch_size: int) -> "DataSet":
        entities = RaggedBatchDict(RaggedBufferF32)
        visible = RaggedBatchDict(RaggedBufferBool)
        actions = RaggedBatchDict(RaggedBufferI64)
        logprobs = RaggedBatchDict(RaggedBufferF32)
        logits = RaggedBatchDict(RaggedBufferF32)
        masks = RaggedActionDict()
        for e in episodes:
            entities.extend(e.entities)
            visible.extend(e.visible)
            actions.extend(e.actions)
            logprobs.extend(e.logprobs)
            masks.extend(e.masks)
            logits.extend(e.logits)

        return cls.from_merged_samples(
            MergedSamples(
                entities=entities,
                visible=visible,
                actions=actions,
                logprobs=logprobs,
                masks=masks,
                logits=logits,
                frames=next(iter(entities.buffers.values())).size0(),
            ),
            batch_size=batch_size,
        )

    @property
    def nbatch(self) -> int:
        return self.frames // self.batch_size

    def batch(
        self, n: int
    ) -> Tuple[
        Dict[str, RaggedBufferF32],
        Dict[str, RaggedBufferBool],
        Dict[str, RaggedBufferI64],
        Dict[str, RaggedBufferF32],
        Dict[str, VecActionMask],
        Optional[Dict[str, RaggedBufferF32]],
    ]:
        if self.permutation is None:
            indices = np.arange(n * self.batch_size, (n + 1) * self.batch_size)
        else:
            indices = self.permutation[n * self.batch_size : (n + 1) * self.batch_size]
        return (
            self.entities[indices],
            self.visible[indices],
            self.actions[indices],
            self.logprobs[indices],
            self.masks[indices],
            self.logits[indices] if self.logits is not None else None,
        )

    def shuffle(self) -> None:
        self.permutation = np.random.permutation(self.frames)

    def deterministic_shuffle(self) -> None:
        stepsize = int(math.sqrt(self.frames))
        perm = np.zeros(self.frames, dtype=np.int64)
        index = 0
        offset = 0
        for i in range(self.frames):
            perm[i] = index
            index += stepsize
            if index >= self.frames:
                offset += 1
                index = offset
        self.permutation = perm


def load_dataset(filepath: str, batch_size: int) -> Tuple[Trace, DataSet, DataSet]:
    trace = Trace.deserialize(open(filepath, "rb").read(), progress_bar=True)

    # episodes = trace.episodes(progress_bar=True)
    # test_episodes = max(len(episodes) // 20, 1) * 2
    # test = episodes[-test_episodes:]
    # train = episodes[:-test_episodes]
    # trainds = DataSet.from_episodes(train, batch_size=batch_size)
    # testds = DataSet.from_episodes(test, batch_size=batch_size)
    train, test = trace.train_test_split(test_frac=0.1, progress_bar=True)
    trainds = DataSet.from_merged_samples(train, batch_size=batch_size)
    testds = DataSet.from_merged_samples(test, batch_size=batch_size)
    print(f"{trainds.frames} training samples")
    print(f"{testds.frames} test samples")
    return trace, trainds, testds


def compute_loss(
    model: RogueNet,
    batch: int,
    ds: DataSet,
    loss_fn: Literal["kl", "mse"],
    tracer: Tracer,
    device: torch.device,
) -> Tuple[torch.Tensor, float]:
    entities, visible, actions, logprobs, masks, logits = ds.batch(batch)
    _, newlogprob, entropy, _, aux, newlogits = model.get_action_and_auxiliary(
        entities=entities,
        visible=visible,
        action_masks=masks,
        prev_actions=actions,
        tracer=tracer,
    )
    loss = torch.tensor(0.0, device=device)
    for actname, target_logprob in logprobs.items():
        # Create normalized distributions
        if loss_fn == "kl":
            if logits is None:
                logprob = newlogprob[actname]
                dist = torch.cat([logprob, (1 - logprob.exp()).log()], dim=1)
                target = torch.tensor(target_logprob.as_array(), device=device)
                target_dist = torch.cat([target.exp(), 1 - target.exp()], dim=1)
                loss += F.kl_div(
                    dist,
                    target_dist,
                )
            else:
                dist = newlogits[actname]
                target_dist = torch.tensor(logits[actname].as_array(), device=device)
                loss += F.kl_div(
                    dist,
                    target_dist.exp(),
                )
        elif loss_fn == "mse":
            logprob = newlogprob[actname]
            target = torch.tensor(target_logprob.as_array().squeeze(-1), device=device)
            loss += F.mse_loss(
                logprob.masked_fill(mask=logprob == float("-inf"), value=0.0),
                target.masked_fill(mask=target == float("-inf"), value=0.0),
            )
    return loss, sum(e.mean().item() for _, e in entropy.items())


def train(
    cfg: Config,
    model: RogueNet,
    trainds: DataSet,
    testds: DataSet,
    device: torch.device,
) -> None:
    tracer = Tracer(cuda=device == "cuda")

    optimizer = AdamW(model.parameters(), lr=cfg.optim.lr)
    for epoch in range(cfg.epochs + 1):
        test_loss = 0.0
        for test_batch in range(testds.nbatch):
            loss, _ = compute_loss(
                model, test_batch, testds, cfg.loss_fn, tracer, device
            )
            test_loss += loss.item()
        test_loss /= testds.nbatch
        print(f"Test loss {test_loss:.4f}")
        if cfg.wandb.track:
            wandb.log(
                {
                    "test_loss": test_loss,
                    "epoch": epoch,
                    "frame": epoch * trainds.frames,
                }
            )
        if epoch == cfg.epochs:
            break

        trainds.shuffle()
        model.train()
        for batch in range(trainds.nbatch):
            frame = batch * trainds.batch_size + epoch * trainds.frames
            if frame % cfg.fast_eval_interval == 0:
                test_loss = 0.0
                for test_batch in range(cfg.fast_eval_samples // testds.batch_size):
                    loss, _ = compute_loss(
                        model,
                        test_batch,
                        testds,
                        cfg.loss_fn,
                        tracer,
                        device,
                    )
                    test_loss += loss.item()
                test_loss /= cfg.fast_eval_samples // testds.batch_size
                print(f"Fast test loss {test_loss:.4f}")
                if cfg.wandb.track:
                    wandb.log(
                        {
                            "fast_test_loss": test_loss,
                            "epoch": epoch,
                            "frame": frame,
                        }
                    )

            optimizer.zero_grad()
            if cfg.optim.anneal_lr:
                frac = 1.0 - frame / (cfg.epochs * trainds.frames)
                lrnow = frac * cfg.optim.lr
                optimizer.param_groups[0]["lr"] = lrnow
            else:
                lrnow = cfg.optim.lr

            loss, entropy = compute_loss(
                model, batch, trainds, cfg.loss_fn, tracer, device
            )
            loss.backward()
            gradnorm = nn.utils.clip_grad_norm_(
                model.parameters(), cfg.optim.max_grad_norm
            )
            optimizer.step()
            if batch % cfg.log_interval == 0:
                print(
                    f"Epoch {epoch}/{cfg.epochs}, Batch {batch}/{trainds.nbatch}, Loss {loss.item():.4f}, Entropy {entropy:.4f}"
                )
            if cfg.wandb.track:
                wandb.log(
                    {
                        "train_loss": loss.item(),
                        "train_entropy": entropy,
                        "gradnorm": gradnorm,
                        "epoch": epoch,
                        "frame": frame,
                        "lr": lrnow,
                    },
                )


@hyperstate.command(Config)
def main(cfg: Config) -> None:
    """Trains a supervised model on samples recorded from an entity-gym environment."""
    trace, traindata, testdata = load_dataset(cfg.dataset_path, cfg.optim.batch_size)
    if testdata.frames < cfg.fast_eval_samples:
        print(
            f"WARNING: fast_eval_samples {cfg.fast_eval_samples} is larger than test dataset {testdata.frames}"
        )
        cfg.fast_eval_samples = testdata.frames
    testdata.deterministic_shuffle()

    if torch.cuda.is_available():
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")
    # TODO: compute input normalization once at the beginning and then freeze it
    model = RogueNet(
        cfg.model,
        obs_space=trace.obs_space,
        action_space=trace.action_space,
    ).to(device)

    if cfg.wandb.track:
        config = asdict(cfg)
        run_name = None
        if os.path.exists("/xprun/info/config.ron"):
            import xprun  # type: ignore

            xp_info = xprun.current_xp()
            config["name"] = xp_info.xp_def.name
            config["base_name"] = xp_info.xp_def.base_name
            config["id"] = xp_info.id
            run_name = xp_info.xp_def.name
        wandb.init(
            project=cfg.wandb.project_name,
            entity=cfg.wandb.entity,
            config=asdict(cfg),
            name=run_name,
        )
        wandb.watch(model)

    assert cfg.loss_fn == "kl" or cfg.loss_fn == "mse"

    train(
        cfg=cfg,
        model=model,
        trainds=traindata,
        testds=testdata,
        device=device,
    )


if __name__ == "__main__":
    main()



import argparse
import heapq
import logging
import math
import random
import threading
import time
from copy import deepcopy
from dataclasses import dataclass, field
from distutils.util import strtobool
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
import optuna
import wandb
import xprun  # type: ignore

logger = logging.getLogger(__name__)


class SamplingStrategy(Enum):
    OMINUS = 0
    POWER_OF_TWO = 1
    LOGUNIFORM = 2
    UNIFORM = 3
    INTUNIFORM = 4


@dataclass
class HyperParam:
    path: str
    sampling_strategy: SamplingStrategy
    min_value: float = -float("inf")
    max_value: float = float("inf")
    constraint: Optional[
        Callable[[Dict[str, float]], Tuple[Optional[float], Optional[float]]]
    ] = None
    transform: Optional[Callable[[Dict[str, float], float], float]] = None

    def suggest(
        self,
        trial: optuna.trial.Trial,
        center: float,
        range: float,
        other_vals: Dict[str, float],
    ) -> Tuple[str, float]:
        min_value, max_value = self.min_value, self.max_value
        if self.constraint is not None:
            _min_value, _max_value = self.constraint(other_vals)
            min_value = max(min_value, _min_value or min_value)
            max_value = min(max_value, _max_value or max_value)
        if self.sampling_strategy == SamplingStrategy.OMINUS:
            min_value = max((1 - center) / range, min_value)
            max_value = min((1 - center) * range, max_value)
            value = 1 - trial.suggest_uniform(self.optuna_name(), min_value, max_value)
        elif self.sampling_strategy == SamplingStrategy.POWER_OF_TWO:
            if center * range < min_value:
                value = min_value
            elif int(center / range) > max_value:
                value = max_value
            else:
                min_value = int(max(center / range, min_value))
                max_value = int(min(center * range, max_value))
                value = 2 ** trial.suggest_int(
                    self.optuna_name(),
                    int(math.log2(min_value)),
                    int(math.log2(max_value)),
                )
        elif self.sampling_strategy == SamplingStrategy.LOGUNIFORM:
            min_value = max(center / range, min_value)
            max_value = min(center * range, max_value)
            value = trial.suggest_loguniform(self.optuna_name(), min_value, max_value)
        elif self.sampling_strategy == SamplingStrategy.UNIFORM:
            min_value = max(center - range, min_value)
            max_value = min(center + range, max_value)
            value = trial.suggest_uniform(f"{self.path}", min_value, max_value)
        elif self.sampling_strategy == SamplingStrategy.INTUNIFORM:
            min_value = int(max(center / range, min_value))
            max_value = int(min(center * range, max_value))
            value = trial.suggest_int(f"{self.path}", min_value, max_value)
        if self.transform is not None:
            value = self.transform(other_vals, value)
        return f"{self.path}={value}", value

    def optuna_name(self) -> str:
        if self.sampling_strategy == SamplingStrategy.OMINUS:
            return f"om_{self.path}"
        elif self.sampling_strategy == SamplingStrategy.POWER_OF_TWO:
            return f"lg_{self.path}"
        else:
            return f"{self.path}"


hyper_params = {
    "learning-rate": HyperParam(
        path="optim.lr",
        sampling_strategy=SamplingStrategy.LOGUNIFORM,
    ),
    "num-envs": HyperParam(
        path="rollout.num_envs",
        sampling_strategy=SamplingStrategy.POWER_OF_TWO,
        min_value=2,
    ),
    "processes": HyperParam(
        path="rollout.processes",
        sampling_strategy=SamplingStrategy.POWER_OF_TWO,
        constraint=lambda x: (1, x["num-envs"]),
    ),
    "d-model": HyperParam(
        path="net.d_model",
        sampling_strategy=SamplingStrategy.POWER_OF_TWO,
    ),
    "n-head": HyperParam(
        path="net.n_head",
        sampling_strategy=SamplingStrategy.POWER_OF_TWO,
        constraint=lambda x: (1, x.get("d-model")),
    ),
    "d-qk": HyperParam(
        path="net.d_qk",
        sampling_strategy=SamplingStrategy.POWER_OF_TWO,
    ),
    "n-layer": HyperParam(
        path="net.n_layer",
        sampling_strategy=SamplingStrategy.INTUNIFORM,
        min_value=1,
    ),
    "num-steps": HyperParam(
        path="rollout.steps",
        sampling_strategy=SamplingStrategy.POWER_OF_TWO,
    ),
    "gamma": HyperParam(
        path="ppo.gamma",
        sampling_strategy=SamplingStrategy.OMINUS,
    ),
    "minibatch-size": HyperParam(
        path="optim.bs",
        sampling_strategy=SamplingStrategy.POWER_OF_TWO,
    ),
    "ent-coef": HyperParam(
        path="ppo.ent_coef",
        sampling_strategy=SamplingStrategy.LOGUNIFORM,
    ),
    "vf-coef": HyperParam(
        path="ppo.vf_coef",
        sampling_strategy=SamplingStrategy.LOGUNIFORM,
    ),
    "max-grad-norm": HyperParam(
        path="optim.max_grad_norm",
        sampling_strategy=SamplingStrategy.LOGUNIFORM,
    ),
}


@dataclass
class Xp:
    xp_def: Any
    trial: int
    trial_manager: "Trial"

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, Xp):
            return self.trial < other.trial
        return NotImplemented

    def __gt__(self, other: Any) -> bool:
        if isinstance(other, Xp):
            return self.trial > other.trial
        return NotImplemented


class HyperOptimizer:
    def __init__(
        self,
        params: List[Tuple[str, float, float]],
        target_metric: str,
        xp_name: str,
        parallelism: int = 6,
        steps: Optional[int] = None,
        time: Optional[int] = None,
        xps_per_trial: int = 1,
        priority: int = 3,
        extra_args: Optional[List[str]] = None,
        average_frac: float = 0.5,
        track: bool = False,
        max_microbatch_size: Optional[int] = None,
        # Only run more than 1 trial if first trial is within variance of best result
        adaptive_trials: bool = False,
        xprun_config: str = "configs/xprun/train.ron",
    ):
        self.xprun = xprun.Client()
        self.wandb = wandb.Api()
        self.trial = 0
        self.time = time
        self.xp_name = xp_name
        xp = self.xprun.build_xpdef(
            xprun_config,
            ignore_dirty=False,
            include_dirty=True,
            verbose=False,
        )
        xp.base_name = self.xp_name
        self.extra_args = extra_args
        self.config = xp
        self.study = optuna.create_study(direction=optuna.study.StudyDirection.MAXIMIZE)
        self.lock = threading.Lock()
        self.cvar = threading.Condition(self.lock)
        self.running_xps = 0
        self.parallelism = parallelism
        self.steps = steps
        self.xps_per_trial = xps_per_trial
        self.best_result: Optional[float] = None
        self.best_result_se: Optional[float] = None
        self.best_config: Optional[str] = None
        self.priority = priority
        self.target_metric = target_metric
        self.average_frac = average_frac
        self.track = track
        self.max_microbatch_size = max_microbatch_size
        self.adaptive_trials = adaptive_trials
        self.pending_xps: List[Xp] = []

        self.last_logged_trial_id = -1
        self.log_in_future: List[Tuple[int, Dict[str, float]]] = []

        self.params = params
        self.steps = steps

        self.done = False

    def run(self, n_trials: int) -> None:
        default_params: Dict[str, float] = {}
        args: Dict[str, float] = {}
        for name, center, _ in self.params:
            if hyper_params[name].sampling_strategy == SamplingStrategy.POWER_OF_TWO:
                center = int(math.log2(center))
            elif hyper_params[name].sampling_strategy == SamplingStrategy.OMINUS:
                center = 1 - center
            oname = hyper_params[name].optuna_name()
            default_params[oname] = center
            args[name] = center
        self.study.enqueue_trial(default_params)

        threading.Thread(target=self.xp_runner).start()

        threads = []
        for trial_id in range(n_trials):
            # Wait until we have a free slot
            with self.lock:
                while len(self.pending_xps) > 0 or self.running_xps >= self.parallelism:
                    self.cvar.wait()
                    logger.debug(
                        f"[main] wake: {len(self.pending_xps)} {self.running_xps}/{self.parallelism}"
                    )
            trial = self.study.ask()
            xp, args = self.sample_xp(trial)
            trial_runner = Trial(
                self,
                self.xps_per_trial,
                xp,
                trial,
                trial_id,
                args,
            )
            logger.debug(f"[main] starting trial {trial_id}")
            thread = threading.Thread(target=trial_runner.run)
            thread.start()
            threads.append(thread)
        for thread in threads:
            thread.join()
        print(f"Best result: {self.best_result}")
        print(f"Best config: {self.best_config}")
        with self.lock:
            self.done = True
            self.cvar.notify_all()

    def xp_runner(self) -> None:
        while True:
            with self.lock:
                while (
                    self.running_xps >= self.parallelism or len(self.pending_xps) == 0
                ):
                    if self.done:
                        return
                    self.cvar.wait()
                    logger.debug(
                        f"[runner] wake: pending={len(self.pending_xps)} running={self.running_xps}/{self.parallelism}"
                    )
                logger.debug(f"[runner] queue: {[xp.trial for xp in self.pending_xps]}")
                next_xp = heapq.heappop(self.pending_xps)
                logger.debug(f"[runner] next: {next_xp.trial}")
                threading.Thread(target=self.run_xp, args=(next_xp,)).start()
                self.running_xps += 1
                self.cvar.notify_all()

    def run_xp(self, xp: Xp) -> None:
        for retry in range(10):
            try:
                self.xprun.run(
                    xp.xp_def, wait=True, priority=self.priority, user="clemens"
                )
            except Exception as e:
                print(f"Failed to run {xp.xp_def.name}: {e}")
                print(f"Retrying in 60 seconds... ({retry})")
                time.sleep(60)
                continue
            break
        else:
            print(f"Failed to run {xp.xp_def.name}")
            return
        while True:
            try:
                self.xprun.block_until_completed(xp.xp_def.name)
                break
            except Exception as e:
                print(f"Failed to block_until_completed {xp.xp_def.name}: {e}")
                print(f"Retrying in 60 seconds... ({retry})")
                time.sleep(60)
                continue
        run = list(
            self.wandb.runs(
                "entity-neural-network/enn-ppo", {"config.name": xp.xp_def.name}
            )
        )[0]
        returns = [
            row[self.target_metric]
            for row in run.scan_history(keys=[self.target_metric])
        ]
        if len(returns) == 0:
            result = -1
        else:
            datapoints = max(1, int(len(returns) * self.average_frac))
            result = np.array(returns[-datapoints:]).mean()
        with self.lock:
            logger.debug(f"[xp] finished {xp.xp_def.name}")
            xp.trial_manager.completed_xps += 1
            xp.trial_manager.results.append(result)
            self.running_xps -= 1
            xp.trial_manager.issue()
            self.cvar.notify_all()

    def base_xp_config(self, trial: int) -> Any:
        xp = deepcopy(self.config)
        xp.name = f"{self.xp_name}-{trial}"
        if self.extra_args is not None:
            xp.containers[0].command.extend(self.extra_args)
        xp.containers[0].command.append(f"total_timesteps={self.steps}")
        if self.time:
            xp.containers[0].command.append(f"max_train_time={self.time}")
        return xp

    def sample_xp(self, trial: optuna.trial.Trial) -> Tuple[Any, Dict[str, float]]:
        xp = self.base_xp_config(self.trial)
        args: Dict[str, float] = {}
        for path, center, range in self.params:
            arg, value = hyper_params[path].suggest(trial, center, range, args)
            args[path] = value
            xp.containers[0].command.append(arg)
        if self.max_microbatch_size is not None:
            minibatch_size = args.get("minibatch-size") or self.max_microbatch_size
            xp.containers[0].command.append(
                f"optim.micro_bs={min(self.max_microbatch_size, minibatch_size)}"
            )
        self.trial += 1
        return xp, args

    def trial_completed(
        self,
        trial: optuna.trial.Trial,
        trial_id: int,
        result: float,
        result_se: float,
        args: Dict[str, float],
        xp: Any,
    ) -> None:
        with self.lock:
            self.study.tell(trial, result)
            print(f"Trial {trial_id}: {result} ± {result_se}")
            if self.track:
                args[self.target_metric] = result
                self.log_in_future.append((trial_id, args))
                # Wandb doesn't like step going backwards
                for _trial_id, _args in sorted(self.log_in_future):
                    if (
                        self.last_logged_trial_id is None
                        or self.last_logged_trial_id + 1 == _trial_id
                    ):
                        wandb.log(
                            _args,
                            step=_trial_id,
                        )
                        self.last_logged_trial_id = _trial_id
            if self.best_result is None or result > self.best_result:
                self.best_result = result
                self.best_result_se = result_se
                command = xp.containers[0].command
                self.best_config = command
                print(f"New best config:\n{' '.join(command)}")


@dataclass
class Trial:
    ctx: HyperOptimizer
    max_xps: int
    xp: Any
    trial: optuna.trial.Trial
    trial_id: int
    args: Dict[str, float]
    results: List[float] = field(default_factory=list)
    completed_xps: int = 0
    authorized: int = 1
    issued: int = 0

    def __post_init__(self) -> None:
        if self.trial_id == 0:
            self.authorized = self.max_xps

    def run(self) -> None:
        with self.ctx.lock:
            while self.completed_xps < self.authorized:
                self.issue()
                logger.debug(
                    f"[trial {self.trial_id}] {self.completed_xps}/{self.issued}/{self.authorized}"
                )
                if self.completed_xps == self.authorized:
                    break

                self.ctx.cvar.wait()

        result, result_se = self.results_mean_se()
        self.ctx.trial_completed(
            self.trial, self.trial_id, result, result_se, self.args, self.xp
        )

    def results_mean_se(self) -> Tuple[float, float]:
        results = np.array(self.results)
        return (
            results.mean(),
            results.std(ddof=1) / np.sqrt(len(results))
            if len(results) > 1
            else self.ctx.best_result_se or 0.0,
        )

    def issue(self) -> None:
        result, result_se = self.results_mean_se()
        if (
            self.ctx.best_result is not None and self.ctx.best_result_se is not None
        ) and result + result_se + self.ctx.best_result_se > self.ctx.best_result:
            self.authorized = max(
                self.authorized,
                min(1 + 2 * self.completed_xps, self.max_xps),
            )

        while self.issued < self.authorized:
            xpid = self.issued
            _xp = deepcopy(self.xp)
            if self.max_xps > 1:
                _xp.containers[0].command.append(f"trial={self.trial_id}")
                _xp.name = f"{_xp.name}-{xpid}"
            heapq.heappush(
                self.ctx.pending_xps,
                Xp(_xp, self.trial_id, self),
            )
            self.issued += 1
            self.ctx.cvar.notify_all()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--n_trials", type=int, default=100)
    parser.add_argument("--params", type=str, nargs="+", default=[])
    parser.add_argument("--parallelism", type=int, default=3)
    parser.add_argument("--steps", type=float)
    parser.add_argument("--time", type=int)  # max training time in seconds
    parser.add_argument("--xps_per_trial", type=int, default=5)
    parser.add_argument("--priority", type=int, default=3)
    parser.add_argument("--run-name", type=str, default=None)
    parser.add_argument("--max-microbatch-size", type=int, default=None)
    parser.add_argument("--adaptive-trials", action="store_true")
    parser.add_argument("--verbose", action="store_true")
    parser.add_argument(
        "--track",
        type=lambda x: bool(strtobool(x)),
        default=False,
        nargs="?",
        const=True,
        help="if toggled, this experiment will be tracked with Weights and Biases",
    )
    parser.add_argument(
        "--average-frac", type=float, default=0.2
    )  # Datapoints from the last average-frac% steps are used to compute final metric
    parser.add_argument("--target-metric", type=str, default="charts/episodic_return")
    parser.add_argument("--xprun-config", type=str, default="configs/xprun/train.ron")
    parser.add_argument("nargs", nargs="*")
    args = parser.parse_args()

    # Configure logging
    logging.basicConfig(level=logging.INFO)
    if args.verbose:
        # Set logging level to DEBUG just for this module
        logger.setLevel(logging.DEBUG)

    run_name = args.run_name or f"optuna-{random.randint(0, 0xffffff):06x}"

    if args.track:
        wandb.init(
            project="enn-ppo-hypertuna",
            entity="entity-neural-network",
            config=vars(args),
            name=args.run_name,
            save_code=True,
        )

    params = []
    for param in args.params:
        path, r = param.split("=")
        center, _range = r.split(":")
        params.append((path, float(center), float(_range)))
    HyperOptimizer(
        params,
        args.target_metric,
        run_name,
        args.parallelism,
        int(args.steps) if args.steps is not None else None,
        args.time,
        xps_per_trial=args.xps_per_trial,
        priority=args.priority,
        extra_args=args.nargs,
        track=args.track,
        max_microbatch_size=args.max_microbatch_size,
        adaptive_trials=args.adaptive_trials,
        xprun_config=args.xprun_config,
        average_frac=args.average_frac,
    ).run(args.n_trials)


"""
poetry run python enn_trainer/enn_trainer/hypertuna.py --track --adaptive-trials --steps=1e9 --time=7200 --n_trials=10 --xps_per_trial=15 --priority=3 --target-metric=episodic_return.mean --parallelism=16 --average-frac=0.05 --max-microbatch-size=2048 \
                                                            --params \
                                                                learning-rate=0.005:100 \
                                                                gamma=0.99:10 \
                                                                minibatch-size=8192:4 \
                                                                ent-coef=0.05:100 \
                                                                vf-coef=1.0:5 \
                                                             -- ppo.anneal_entropy=true rollout.num_envs=256 rollout.steps=128 net.d_model=128 net.n_head=2 net.n_layer=2 env.id=MultiSnake track=true env.kwargs='{"num_snakes": 2, "max_snake_length": 11}' rollout.processes=16 optim.max_grad_norm=10 ppo.anneal_entropy=True net.relpos_encoding='(extent: [10, 10], position_features: ["x", "y"])'
"""



from entity_gym.examples import ENV_REGISTRY
from hyperstate import StateManager
from rogue_net.relpos_encoding import RelposEncodingConfig
from rogue_net.rogue_net import RogueNetConfig

from enn_trainer.config import RolloutConfig
from enn_trainer.train import (
    EnvConfig,
    OptimizerConfig,
    PPOConfig,
    State,
    TrainConfig,
    init_train_state,
    train,
)


def _train(cfg: TrainConfig) -> float:
    sm = StateManager(TrainConfig, State, init_train_state, None)
    sm._config = cfg
    return train(sm, ENV_REGISTRY[cfg.env.id])


def test_multi_armed_bandit() -> None:
    cfg = TrainConfig(
        total_timesteps=500,
        cuda=False,
        ppo=PPOConfig(ent_coef=0.0, gamma=0.5, anneal_entropy=False),
        env=EnvConfig(id="MultiArmedBandit"),
        rollout=RolloutConfig(steps=16, processes=2, num_envs=4),
        net=RogueNetConfig(n_layer=0, d_model=16),
        optim=OptimizerConfig(lr=0.05, bs=16, update_epochs=4),
    )
    meanrew = _train(cfg)
    print(f"Final mean reward: {meanrew}")
    assert meanrew > 0.99 / 32


def test_minefield() -> None:
    cfg = TrainConfig(
        total_timesteps=256,
        cuda=False,
        net=RogueNetConfig(d_model=16),
        env=EnvConfig(id="Minefield"),
        rollout=RolloutConfig(steps=16, num_envs=4),
        optim=OptimizerConfig(bs=64),
        ppo=PPOConfig(),
    )
    meanrew = _train(cfg)
    print(f"Final mean reward: {meanrew}")
    assert meanrew >= 0.0


def test_multi_snake() -> None:
    cfg = TrainConfig(
        total_timesteps=256,
        cuda=False,
        net=RogueNetConfig(d_model=16),
        env=EnvConfig(id="MultiSnake"),
        rollout=RolloutConfig(steps=16, num_envs=4),
        optim=OptimizerConfig(bs=64),
        ppo=PPOConfig(),
    )
    meanrew = _train(cfg)
    print(f"Final mean reward: {meanrew}")
    assert meanrew >= 0.0


def test_not_hotdog() -> None:
    cfg = TrainConfig(
        total_timesteps=1000,
        cuda=False,
        net=RogueNetConfig(d_model=16, n_layer=1),
        env=EnvConfig(id="NotHotdog"),
        rollout=RolloutConfig(steps=16, num_envs=8),
        optim=OptimizerConfig(bs=16, lr=0.005),
        ppo=PPOConfig(ent_coef=0.0, gamma=0.5),
    )
    meanrew = _train(cfg)
    print(f"Final mean reward: {meanrew}")
    assert meanrew >= 0.99


def test_masked_count() -> None:
    cfg = TrainConfig(
        total_timesteps=2000,
        cuda=False,
        net=RogueNetConfig(d_model=16, n_layer=1),
        env=EnvConfig(id="Count", kwargs='{"masked_choices": 2}'),
        rollout=RolloutConfig(steps=1, num_envs=16),
        optim=OptimizerConfig(bs=16, lr=0.01),
        ppo=PPOConfig(),
    )
    meanrw = _train(cfg)
    print(f"Final mean reward: {meanrw}")
    assert meanrw >= 0.99


def test_pick_matching_balls() -> None:
    cfg = TrainConfig(
        total_timesteps=256,
        cuda=False,
        net=RogueNetConfig(d_model=16),
        env=EnvConfig(id="PickMatchingBalls"),
        rollout=RolloutConfig(steps=16, num_envs=4),
        optim=OptimizerConfig(bs=64),
        ppo=PPOConfig(),
    )
    meanrew = _train(cfg)
    print(f"Final mean reward: {meanrew}")
    assert meanrew >= 0.0


def test_cherry_pick() -> None:
    cfg = TrainConfig(
        total_timesteps=256,
        cuda=False,
        net=RogueNetConfig(d_model=16),
        env=EnvConfig(id="CherryPick"),
        rollout=RolloutConfig(steps=16),
        optim=OptimizerConfig(bs=64),
        ppo=PPOConfig(),
    )
    meanrew = _train(cfg)
    print(f"Final mean reward: {meanrew}")


def test_relpos_encoding() -> None:
    cfg = TrainConfig(
        total_timesteps=10000,
        cuda=False,
        net=RogueNetConfig(
            d_model=16,
            n_layer=2,
            relpos_encoding=RelposEncodingConfig(
                extent=[1, 1],
                position_features=["x", "y"],
                per_entity_values=True,
            ),
        ),
        env=EnvConfig(id="FloorIsLava"),
        rollout=RolloutConfig(steps=2, num_envs=64),
        optim=OptimizerConfig(bs=32, lr=0.03),
        ppo=PPOConfig(
            ent_coef=1.0,
            anneal_entropy=True,
        ),
    )
    meanrew = _train(cfg)
    print(f"Final mean reward: {meanrew}")
    assert meanrew >= 0.95

    cfg.net.relpos_encoding = None
    meanrew = _train(cfg)
    print(f"Final mean reward: {meanrew}")
    assert meanrew < 0.3


def test_asymmetric_relpos_encoding() -> None:
    cfg = TrainConfig(
        total_timesteps=3000,
        cuda=False,
        net=RogueNetConfig(
            d_model=16,
            n_layer=2,
            relpos_encoding=RelposEncodingConfig(
                extent=[5, 1],
                position_features=["x", "y"],
                per_entity_values=True,
            ),
        ),
        env=EnvConfig(id="FloorIsLava"),
        rollout=RolloutConfig(steps=1, num_envs=64),
        optim=OptimizerConfig(bs=16, lr=0.02),
        ppo=PPOConfig(
            ent_coef=0.1,
            anneal_entropy=True,
        ),
    )

    meanrew = _train(cfg)
    print(f"Final mean reward: {meanrew}")
    assert meanrew >= 0.15


def test_rock_paper_scissors() -> None:
    cfg = TrainConfig(
        total_timesteps=4000,
        cuda=False,
        net=RogueNetConfig(d_model=16, n_layer=2),
        env=EnvConfig(id="RockPaperScissors"),
        rollout=RolloutConfig(steps=1, num_envs=256),
        optim=OptimizerConfig(bs=64, lr=0.01),
        ppo=PPOConfig(ent_coef=0.1),
    )
    meanrew = _train(cfg)
    print(f"Final mean reward: {meanrew}")
    assert 0.8 <= meanrew <= 1.1

    cfg.env.kwargs = '{"cheat": true}'
    meanrew = _train(cfg)
    print(f"Final mean reward: {meanrew}")
    assert meanrew > 1.9
